#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Code for the structural segmentation evaluation in our ISMIR 2014 paper
Karen Ullrich, Jan Schlüter, Thomas Grill (2014): "Boundary Detection in Music
Structure Analysis using Convolutional Neural Networks".

For usage information, call without any parameters.

Author: Jan Schlüter; uses parts of a library by CP/JKU and OFAI (included)
"""

import sys
import os
import numpy as np
import itertools

import madmom.evaluation
import madmom.evaluation.onsets

def print_usage():
	print 'eval.py: Code for the structural segmentation evaluation in'
	print '  our ISMIR 2014 paper "Boundary Detection in Music Structure '
	print '  Analysis using Convolutional Neural Networks.'
	print 'Usage: %s TOL PREDICTIONS TRUTH' % sys.argv[0]
	print '  TOL: Temporal tolerance to apply, we used 0.5 and 3.0 (seconds)'
	print '  PREDICTIONS: Either a single text file of boundary predictions;'
	print '    or a directory of such files (and nothing else)'
	print '  TRUTH: Either a single text file of boundary annotations;'
	print '    or a directory of such files (and nothing else)'
	print 'The annotations and predictions must be in one of these formats:'
	print '1) One line per boundary giving the boundary positions in seconds'
	print '2) One line per segment giving the segment start and end positions'
	print '   in seconds, deliminated by whitespace'
	print 'Each line can be followed by whitespace and a textual label.'

def evaluate_multiple_files(predictions, truths, tolerance, mode='mean'):
	if mode == 'sum':
		se = madmom.evaluation.SumEvaluation()
		for pred, truth in itertools.izip(predictions, truths):
			se += madmom.evaluation.onsets.OnsetEvaluation(pred, truth, tolerance)
	elif mode == 'mean':
		se = madmom.evaluation.MeanEvaluation()
		for pred, truth in itertools.izip(predictions, truths):
			se.append(madmom.evaluation.onsets.OnsetEvaluation(pred, truth, tolerance))
	return se

def read_file(filename):
	bounds = set()
	with open(filename, 'rb') as f:
		for l in f:
			l = l.split(None, 2)  # split at white space, at most twice
			if l:
				# if the line is not empty, we expect the first token to
				# be a floating point number. otherwise we will fail to
				# make sure the user notices.
				bounds.add(float(l[0]))
			try:
				# we try to interpret the second token as a floating point
				# number as well. if there is no second token or it is not
				# numeric, we just ignore it.
				bounds.add(float(l[1]))
			except Exception:
				pass
	return sorted(bounds)

def read_files(path):
	try:
		return [read_file(path)]
	except IOError:
		return [read_file(os.path.join(path, filename))
				for filename in sorted(os.listdir(path))
				if os.path.isfile(os.path.join(path, filename))]

def main():
	if len(sys.argv) != 4:
		print_usage()
		return
	
	# 'parse' command line
	tol, preds, truths = sys.argv[1:]
	tol = float(tol)
	
	# read files
	preds = read_files(preds)
	truths = read_files(truths)
	if len(preds) != len(truths):
		print "Found %d prediction files, but %d ground truth files. Aborting." % (len(preds), len(truths))
		return
	print "Evaluating on %d file%s." % (len(preds), 's'*(len(preds) > 1))
	
	# evaluate
	result = evaluate_multiple_files(preds, truths, tolerance=tol, mode='mean')
	
	# print
	print "Tolerance: +/- %g seconds" % tol
	print "F-measure: %.2f%%" % (result.fmeasure * 100)
	print "Precision: %.2f%%" % (result.precision * 100)
	print "Recall:    %.2f%%" % (result.recall * 100)

if __name__=="__main__":
	main()

